from __future__ import annotations

import shutil
from pathlib import Path

import requests
from bs4 import BeautifulSoup

UA = {
    "User-Agent": "Mozilla/5.0 (Macintosh Intel Mac OS X 10_13_5) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/67.0.3396.99 Safari/537.36",
}


class Download:
    def __init__(self, site: str) -> None:
        """Init."""
        self.site = site
        self.save_dir = Path.cwd() / self.site

    def get_all_pages(self) -> None:
        """Get pages till a page without an article is shown."""
        if self.site == "dlsite":
            self.get_dlsite_pages()

    def get_dlsite_pages(self) -> None:
        root = (
            "https://www.dlsite.com/maniax/fsr/=/language/jp/"
            "sex_category%5B0%5D/male/work_category%5B0%5D/doujin/"
            "order%5B0%5D/release_d/work_type%5B0%5D/SOU/"
            "work_type_name%5B0%5D/%E3%83%9C%E3%82%A4%E3%82%B9%E3%83%BBASMR/"
            "per_page/100/page/{}"
        )

        shutil.rmtree(self.save_dir, ignore_errors=True)
        self.save_dir.mkdir(parents=True, exist_ok=True)

        pagenation = 0
        while True:
            pagenation += 1
            filename = f"{pagenation:05}.html"
            url = root.format(pagenation)
            print(f"\33[2K\rnow: {filename}", end="", flush=True)  # noqa: T201
            page_source = requests.get(url, headers=UA, timeout=10).text
            if self.__check_work(page_source):
                self.__save_file(page_source, filename)
            else:
                break

    def __check_work(self, page: str) -> bool:
        """Judge if articles exists in a source."""
        bs = BeautifulSoup(page, "lxml")
        return bs is not None and bs.select_one("li[class=search_result_img_box_inner]") is not None

    def __save_file(self, source: str, filename: str) -> None:
        """Save a file."""
        save_path = Path(self.save_dir) / filename
        print(source, file=save_path.open("w"))



from __future__ import annotations

import logging
from sqlite3 import Error, connect
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .parser import DlsiteDict


class DojinvoiceDatabase:
    def __init__(self, db_filepath: str, log: str = ".log") -> None:
        """Init."""
        self.db_filepath = db_filepath
        f = logging.FileHandler(log)
        f.setLevel(logging.DEBUG)
        self.__init_database()

    def __init_database(self) -> None:
        """Create a db file and tables if not exists."""
        connect(self.db_filepath)

    def __connect_db(self, sql: str, p: list[tuple[Any, ...]] | None = None) -> None:
        """Connect to the database and execute the SQL."""
        conn = None
        try:
            conn = connect(self.db_filepath)

            c = conn.cursor()
            if p is None:
                c.execute(sql)
            else:
                c.executemany(sql, p)

            conn.commit()
        except Error as e:
            msg = str(e)
            if msg[0:5] == "table" and msg[-14:] == "already exists":
                pass
            else:
                logging.exception("[DB Error]: {e}")
                logging.exception("- [SQL]: {sql}")
                logging.exception("- [Return]: {p}")
        finally:
            if conn:
                conn.close()

    def get_work_ids(self) -> list[str]:
        conn = connect(self.db_filepath)
        c = conn.cursor().execute("select work_id from work")
        res = [i[0] for i in c]
        conn.close()
        return res

    def __create_table(self, schema: str) -> None:
        self.__connect_db("create table " + schema)

    def create_tables(self) -> None:
        """Create the tables."""

        self.__create_table(
            """work (
                    work_id text primary key,
                    detail_link text not null,
                    title text not null,
                    circle text not null,
                    circle_link text not null,
                    category text not null,
                    sale_date integer not null,
                    age_zone text not null,
                    file_format text not null,
                    file_size text not null,
                    description str not null,
                    monopoly integer not null,
                    price integer not null
                ) """,
        )
        self.__create_table(
            """option (
                    work_id text primary key,
                    thumbnail text,
                    cien_link text,
                    series text,
                    chobit_link text,
                    sales integer,
                    favorites integer,
                    trial_link text,
                    trial_size integer,
                    rating real
                ) """,
        )
        self.__create_table(
            """writer (
                    work_id text,
                    writer text,
                    primary key (work_id, writer)
                ) """,
        )
        self.__create_table(
            """scenario (
                    work_id text,
                    scenario text,
                    primary key (work_id, scenario)
                ) """,
        )
        self.__create_table(
            """illustrator (
                    work_id text,
                    illustrator text,
                    primary key (work_id, illustrator)
                ) """,
        )
        self.__create_table(
            """voice (
                    work_id text,
                    voice text,
                    primary key (work_id, voice)
                ) """,
        )
        self.__create_table(
            """musician (
                    work_id text,
                    musician text,
                    primary key (work_id, musician)
                ) """,
        )
        self.__create_table(
            """genre (
                    work_id text,
                    genre text,
                    primary key (work_id, genre)
                ) """,
        )

    def push(self, data: list[DlsiteDict]) -> None:
        """Insert data to the database."""
        work_data: list[tuple[Any, ...]] = []
        option_data: list[tuple[Any, ...]] = []
        writer_data: list[tuple[Any, ...]] = []
        scenario_data: list[tuple[Any, ...]] = []
        illustrator_data: list[tuple[Any, ...]] = []
        voice_data: list[tuple[Any, ...]] = []
        musician_data: list[tuple[Any, ...]] = []
        genre_data: list[tuple[Any, ...]] = []
        for datum in data:
            work_data.append(
                (
                    datum["work_id"],
                    datum["detail_link"],
                    datum["title"],
                    datum["circle"],
                    datum["circle_link"],
                    datum["category"],
                    datum["sale_date"],
                    datum["age_zone"],
                    datum["file_format"],
                    datum["file_size"],
                    datum["description"],
                    datum["monopoly"],
                    datum["price"],
                ),
            )

            option_data.append(
                (
                    datum["work_id"],
                    datum["thumbnail"],
                    datum["cien_link"],
                    datum["series"],
                    datum["chobit_link"],
                    datum["sales"],
                    datum["favorites"],
                    datum["trial_link"],
                    datum["trial_size"],
                    datum["rating"],
                ),
            )

            if datum["writers"] is not None:
                writer_data.extend([(datum["work_id"], writer) for writer in datum["writers"]])

            if datum["voices"] is not None:
                voice_data.extend([(datum["work_id"], voice) for voice in datum["voices"]])

            if datum["scenarios"] is not None:
                scenario_data.extend([(datum["work_id"], scenario) for scenario in datum["scenarios"]])

            if datum["illustrators"] is not None:
                illustrator_data.extend([(datum["work_id"], illustrator) for illustrator in datum["illustrators"]])

            if datum["musicians"] is not None:
                musician_data.extend([(datum["work_id"], musician) for musician in datum["musicians"]])

            if datum["genres"] is not None:
                genre_data.extend([(datum["work_id"], genre) for genre in datum["genres"]])

        query_lists: list[tuple[str, list[tuple[Any, ...]]]] = [
            ("insert into work values (?,?,?,?,?,?,?,?,?,?,?,?,?)", work_data),
            ("insert into option values (?,?,?,?,?,?,?,?,?,?)", option_data),
            ("insert into writer values (?,?)", writer_data),
            ("insert into scenario values (?,?)", scenario_data),
            ("insert into illustrator values (?,?)", illustrator_data),
            ("insert into voice values (?,?)", voice_data),
            ("insert into musician values (?,?)", musician_data),
            ("insert into genre values (?,?)", genre_data),
        ]

        for query, data_list in query_lists:
            unique_data_list = list(set(data_list))
            if len(unique_data_list) != 0:
                self.__connect_db(query, unique_data_list)



from __future__ import annotations

import urllib.parse
from datetime import datetime, timezone
from pathlib import Path
from random import uniform
from time import sleep
from typing import TypedDict, cast

from bs4 import BeautifulSoup
from bs4.element import Tag
from humanfriendly import parse_size
from selenium import webdriver
from selenium.common.exceptions import TimeoutException


class UnknownSiteError(Exception):
    pass


class DlsiteDict(TypedDict):
    work_id: str
    detail_link: str
    title: str
    circle: str
    circle_link: str
    sale_date: int
    age_zone: str
    category: str
    file_format: str
    file_size: int
    description: str
    monopoly: bool
    price: int
    thumbnail: str | None
    cien_link: str | None
    series: str | None
    writers: list[str] | None
    scenarios: list[str] | None
    illustrators: list[str] | None
    voices: list[str] | None
    musicians: list[str] | None
    chobit_link: str | None
    sales: int | None
    favorites: int | None
    trial_link: str | None
    trial_size: int | None
    rating: float | None
    genres: list[str] | None


class DmmDict(TypedDict):
    work_id: str
    detail_link: str
    title: str
    thumbnail: str
    samples: list[str | None]
    circle: str
    circle_link: str
    sale_date: int
    use_limit: str | None
    voice_num: str
    file_size: str
    series: str | None
    subject: str
    genre: list[str]
    trial_link: str | None
    description: str
    rating: float | None
    sales: int | None
    favorites: int | None
    price: int


UA = {
    "User-Agent": "Mozilla/5.0 (Macintosh Intel Mac OS X 10_13_5) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/67.0.3396.99 Safari/537.36",
}


class Parser:
    def __init__(self, site: str, exclude_ids: list[str] | None = None) -> None:
        """Init."""
        self.site = site
        self.exclude_ids = exclude_ids

        options = webdriver.ChromeOptions()
        options.add_argument("--headless")
        options.add_argument("--disable-gpu")
        options.add_argument("--disable-extensions")
        options.add_argument("--disable-infobars")
        options.add_argument("--disable-dev-shm-usage")
        options.add_argument("--disable-browser-side-navigation")
        options.add_argument('--proxy-server="direct://"')
        options.add_argument("--proxy-bypass-list=*")
        options.add_argument("--start-maximized")
        options.add_argument("--user-agent={}".format(UA["User-Agent"]))
        options.add_experimental_option("excludeSwitches", ["enable-automation"])
        options.add_experimental_option("useAutomationExtension", False)  # noqa: FBT003

        print("Preparing for headless chrome...", end="", flush=True)  # noqa: T201

        self.driver = webdriver.Chrome(options=options)
        self.driver.set_page_load_timeout(60)
        self.driver.get("https://www.dlsite.com")
        self.driver.add_cookie({"name": "locale", "value": "ja-jp"})
        self.driver.add_cookie({"name": "adultchecked", "value": "1"})
        self.driver.get("https://www.dlsite.com/maniax/work/=/product_id/RJ305341.html")

    def parse(self, path: Path, page_idx: int = 0) -> list[DlsiteDict] | list[DmmDict]:
        """Extract required information from the page sources and scrape it."""
        self.page_idx = page_idx
        if self.site == "dlsite":
            return self.__parse_dlsite_pages(path)
        raise UnknownSiteError("Unknown Site: %s" % self.site)

    def __parse_dlsite_pages(self, path: Path) -> list[DlsiteDict]:  # noqa: PLR0915
        res = []
        bs = BeautifulSoup(path.open().read(), "lxml")
        work_links = [
            str(a.get("href")) for a in bs.select("a[class=work_thumb_inner]") if isinstance(a.get("href"), str)
        ]
        thumb_links = [
            str(img.get("src") or img.get("data-src"))
            for img in bs.select("img[class=lazy]")
            if isinstance(img.get("src"), str) or isinstance(img.get("data-src"), str)
        ]
        if len(work_links) != len(thumb_links):
            msg = f"number of works != number of thumbnails: ({len(work_links)} != {len(thumb_links)})"
            raise ValueError(msg)
        for idx, (work_link, thumb_link) in enumerate(zip(work_links, thumb_links)):
            print(  # noqa: T201
                f"\33[2K\r{100 * self.page_idx + idx + 1}: {work_link}",
                end="",
            )
            o = urllib.parse.urlparse(work_link)
            work_id = Path(o.path).stem
            if self.exclude_ids and work_id in self.exclude_ids:
                continue
            data = cast("DlsiteDict", {})

            for _ in range(5):
                try:
                    self.driver.get(work_link)
                    break
                except TimeoutException:
                    pass
            else:
                raise ConnectionError

            bs = BeautifulSoup(self.driver.page_source, "lxml")
            if bs is None:
                continue
            title = bs.select_one("h1[id=work_name]")
            if title is None:
                print("404 skipped:", work_link)  # noqa: T201
                continue

            data["work_id"] = work_id

            data["detail_link"] = work_link
            data["title"] = title.string if title.string is not None else ""

            data["thumbnail"] = f"https:{thumb_link}"

            circle_elm = bs.select_one("span[class=maker_name] > a")
            data["circle"] = circle_elm.string if circle_elm is not None and circle_elm.string else ""

            circle_link_elm = bs.select_one("span[class=maker_name] a")
            data["circle_link"] = str(circle_link_elm.get("href", "")) if circle_link_elm is not None else ""

            cien_elm = bs.select_one("div[class=link_cien] a")
            data["cien_link"] = str(cien_elm.get("href", "")) if cien_elm is not None else None

            info_table = {
                str(tr.th.string): tr.td
                for tr in bs.select("table[id=work_outline] > tbody > tr")
                if isinstance(tr, Tag)
                and isinstance(tr.th, Tag)
                and isinstance(tr.th.string, str)
                and tr.td is not None
            }

            sale_date_elm = info_table["販売日"].select_one("a") if "販売日" in info_table else None
            sale_date = (
                str(sale_date_elm.string) if sale_date_elm is not None and sale_date_elm.string is not None else ""
            )

            if "時" in sale_date:
                data["sale_date"] = int(
                    datetime.strptime(sale_date, "%Y年%m月%d日 %H時").astimezone(timezone.utc).timestamp(),
                )
            else:
                data["sale_date"] = int(
                    datetime.strptime(sale_date, "%Y年%m月%d日").astimezone(timezone.utc).timestamp(),
                )

            category_elm = info_table["作品形式"].select_one("a") if "作品形式" in info_table else None
            data["category"] = (
                category_elm.string if category_elm is not None and category_elm.string is not None else ""
            )

            series_elm = info_table["シリーズ名"].select_one("a") if "シリーズ名" in info_table else None
            data["series"] = series_elm.string if series_elm is not None and series_elm.string is not None else ""

            data["writers"] = (
                [writer.string for writer in info_table["作者"].select("a") if writer.string is not None]
                if "作者" in info_table
                else None
            )
            data["scenarios"] = (
                [scenario.string for scenario in info_table["シナリオ"].select("a") if scenario.string is not None]
                if "シナリオ" in info_table
                else None
            )
            data["illustrators"] = (
                [
                    illustrators.string
                    for illustrators in info_table["イラスト"].select("a")
                    if illustrators.string is not None
                ]
                if "イラスト" in info_table
                else None
            )
            data["voices"] = [_.string for _ in info_table["声優"].select("a")] if "声優" in info_table else None
            data["musicians"] = [_.string for _ in info_table["音楽"].select("a")] if "音楽" in info_table else None
            data["age_zone"] = ",".join(
                [age_zone.string for age_zone in info_table["年齢指定"].select("span") if age_zone.string is not None],
            )
            data["file_format"] = info_table["ファイル形式"].get_text()
            data["genres"] = (
                [genre.string for genre in info_table["ジャンル"].select("a") if genre.string is not None]
                if "ジャンル" in info_table
                else None
            )
            file_size_elm = info_table["ファイル容量"].select_one("div[class=main_genre]")
            file_size = file_size_elm.string if file_size_elm is not None and file_size_elm.string is not None else ""
            data["file_size"] = parse_size(file_size.replace("計", "").replace("総", ""))

            trial_elm = bs.select_one("div[class*=trial_download] > ul > li")
            data["trial_link"], data["trial_size"] = None, None
            if trial_elm is not None:
                trial_link_elm = trial_elm.select_one("a[class=btn_trial]")
                data["trial_link"] = (
                    f"https:{trial_link_elm.get('href')}"
                    if trial_link_elm is not None and isinstance(trial_link_elm.get("href"), str)
                    else None
                )
                trial_size_elm = trial_elm.select_one("span")
                data["trial_size"] = (
                    parse_size(trial_size_elm.string.replace("(", "").replace(")", ""))
                    if trial_size_elm is not None and trial_size_elm.string is not None
                    else None
                )

            description_elm = bs.select_one("div[itemprop=description]")
            data["description"] = description_elm.get_text() if description_elm is not None else ""

            data["monopoly"] = bs.select_one("span[title=DLsite専売]") is not None

            rate_elm = bs.select_one("span[class='point average_count']")
            data["rating"] = float(rate_elm.string) if rate_elm is not None and rate_elm.string is not None else None

            sales_elm = bs.select_one("dd[class=point]")
            data["sales"] = (
                int(sales_elm.string.replace(",", ""))
                if sales_elm is not None and sales_elm.string is not None
                else None
            )

            favorites_elm = bs.select_one("dd[class=position_fix]")
            data["favorites"] = (
                int(favorites_elm.string.replace(",", ""))
                if favorites_elm is not None and favorites_elm.string is not None
                else None
            )

            price_elm = bs.select_one("div[class=price]")
            data["price"] = (
                int(price_elm.text.replace(",", "").replace("円", ""))
                if price_elm is not None and price_elm.string is not None
                else 0
            )

            chobit_elm = bs.select_one("div[class='work_parts type_chobit']>iframe")
            data["chobit_link"] = (
                str(chobit_elm.get("src"))
                if chobit_elm is not None and isinstance(chobit_elm.get("src"), str)
                else None
            )

            sleep(uniform(0.1, 1.0))  # noqa: S311

            # breakpoint()  # noqa: ERA001

            res.append(data)

        return res



from __future__ import annotations

from pathlib import Path
from typing import List, cast

from .database import DojinvoiceDatabase
from .download import Download
from .parser import DlsiteDict, Parser


def get_filepaths(dirpath: str) -> list[Path]:
    """Get the file paths in the specified directory."""
    path = Path.cwd() / dirpath
    return sorted(path.glob("*"))


def main() -> None:
    """Main!!!"""
    if input("Download pages? >> ") == "y":
        Download("dlsite").get_all_pages()

    db = DojinvoiceDatabase("dojinvoice.db")
    db.create_tables()

    exclude_ids: list[str] = db.get_work_ids()
    if len(exclude_ids) > 0:
        print(len(exclude_ids), "ids was committed to existed db!")  # noqa: T201

    for site in ("dlsite",):
        print("site:", site)  # noqa: T201
        p = Parser(site, exclude_ids)
        targets = get_filepaths(site)
        print(len(targets), "lists of works is found!")  # noqa: T201
        for page_idx, path in enumerate(targets):
            print(f"\033[2K\r\033[31mNow: {path}\033[0m")  # noqa: T201
            d = p.parse(path, page_idx)
            print(" =>committing to DB...", end="")  # noqa: T201
            db.push(cast(List[DlsiteDict], d))

        print("DONE:", site)  # noqa: T201


if __name__ == "__main__":
    main()
